/*
 * lf2.c, LEAF LF2 plug-in for GIMP
 * Copyright (C) 1997 TF <tf@denpa.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * lf2.c supports the LF2 format image files(INDEX/INDEXA). This format
 * is used in the LEAF's game 'To Heart'.
 *
 * Changes:
 *   11/15/1997 supported offset in output images.
 *   11/07/1977 supported output.
 *   11/06/1997 original code was written (read only).
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <setjmp.h>
#include "gtk/gtk.h"
#include "libgimp/gimp.h"

#define LF2_R 0
#define LF2_G 1
#define LF2_B 2

#define LF2_ENTRY_WIDTH 64
#define LF2_BUFLEN 64

#define LF2_COLOR_NUM 256

#define RINGBUF_SIZE 0x1000    /* size of ring buffer */
#define F 18    /* longest match length */
#define NIL RINGBUF_SIZE    /* end of tree */

typedef struct {
  int width;
  int height;
  int xoffset;
  int yoffset;
  int color_num;
  int unknown;         /* unknown, yet */
  int transparent;     /* the palette num. of transparent color */
  int size;            /* the size of image */
  int palette[LF2_COLOR_NUM][3];
  int *image;
  int *encoded_data;
  int encoded_size;
} LF2;

/*
 * plug-in interfaces
*/
static void query(void);
static void run(char *, int, GParam *, int *, GParam **);
static gint32 load_image (char *);
gint save_image(char *, gint32, gint32);

/*
 * dialog and callbacks
 */
static gint save_dialog(void);
static void save_ok_callback(GtkWidget *, gpointer);
static void save_close_callback(GtkWidget *, gpointer);
static void lf2_xentry_callback (GtkWidget *, gpointer);
static void lf2_yentry_callback (GtkWidget *, gpointer);

/*
 * for read
 */
static LF2 *read_lf2(FILE *);
static LF2 *init_lf2(void);
static void free_lf2(LF2 *);
static int read_magic(FILE *);
static void read_unknown(FILE *, LF2 *);
static void read_size(FILE *, LF2 *);
static void read_palette(FILE *, LF2 *);
static void read_color_num(FILE *, LF2 *);
static void read_transparent(FILE *, LF2 *);
static void skip_padding(FILE *, int);
static int extract_body(FILE *, LF2 *);
static void lf2_expand_body(LF2 *, GPixelRgn *);

/*
 * for output
 */
static int output_lf2(LF2 *, FILE *);
static void output_word(int, FILE *);
void init_tree(void);
void insert_node(int);
void delete_node(int);
void lf2_encode(guchar *, LF2 *);
static void lf2_set_cmap(LF2 *, int, guchar *);

/*
 * for encoding
 */
unsigned long outcount = 0;  /* counter for output size */
unsigned char text[RINGBUF_SIZE+F-1]; /* text buffer */
int dad[RINGBUF_SIZE+1];
int lson[RINGBUF_SIZE+1];
int rson[RINGBUF_SIZE+257];  /* tree */

int matchpos;  /* the position of longest match pattern */
int matchlen;  /* the length of longest match pattern */

/*
 * plug-in interfaces
*/
GPlugInInfo PLUG_IN_INFO = {
  NULL,    /* init_proc */
  NULL,    /* quit_proc */
  query,   /* query_proc */
  run,     /* run_proc */
};

typedef struct {
  gint run;
} LF2SaveInterface;

static LF2SaveInterface lf2sint = {
  FALSE   /*  run  */
};

typedef struct {
  int xoffset;
  int yoffset;
} LF2SaveVals;

static LF2SaveVals lf2svals = {
  0,    /* xoffset */
  0,    /* yoffset */
};

MAIN ();

static void query (void)
{
  static GParamDef load_args[] = {
    { PARAM_INT32, "run_mode", "Interactive, non-interactive" },
    { PARAM_STRING, "filename", "The name of the file to load" },
    { PARAM_STRING, "raw_filename", "The name entered" },
  };

  static GParamDef load_return_vals[] = {
    { PARAM_IMAGE, "image", "Output image" },
  };

  static GParamDef save_args[] = {
    { PARAM_INT32, "run_mode", "Interactive, non-interactive" },
    { PARAM_IMAGE, "image", "Input image" },
    { PARAM_DRAWABLE, "drawable", "Drawable to save" },
    { PARAM_STRING, "filename", "The name of the file to save the image in" },
    { PARAM_STRING, "raw_filename", "The name entered" },
  };

  static int nload_args = sizeof (load_args) / sizeof (load_args[0]);
  static int nload_return_vals = sizeof (load_return_vals) / sizeof (load_return_vals[0]);
  static int nsave_args = sizeof (save_args) / sizeof (save_args[0]);

  gimp_install_procedure ("file_lf2_load",
                          "loads files of the lf2 file format",
                          "FIXME: write help for lf2_load",
                          "FUJIWARA Teruyoshi",
                          "FUJIWARA Teruyoshi",
                          "1997",
                          "<Load>/LF2",
			  NULL,
                          PROC_PLUG_IN,
                          nload_args, nload_return_vals,
                          load_args, load_return_vals);

  gimp_install_procedure ("file_lf2_save",
                          "saves files in lf2 file format",
                          "FIXME: write help for lf2_save",
                          "FUJIWARA Teruyoshi",
                          "FUJIWARA Teruyoshi",
                          "1997",
                          "<Save>/LF2",
			  "INDEXED*",
                          PROC_PLUG_IN,
                          nsave_args, 0,
                          save_args, NULL);

  gimp_register_magic_load_handler ("file_lf2_load", "lf2", "", "0,string,LEAF256\0");
  gimp_register_save_handler ("file_lf2_save", "lf2", "");
}

static void run (char *name, int nparams, GParam *param, int *nreturn_vals, GParam **return_vals)
{
  static GParam values[2];
  GRunModeType run_mode;
  gint32 image_ID;
  
  run_mode = param[0].data.d_int32;
  
  *nreturn_vals = 1;
  *return_vals = values;
  
  values[0].type = PARAM_STATUS;
  values[0].data.d_status = STATUS_CALLING_ERROR;
  
  if (!strcmp(name, "file_lf2_load")) {
    image_ID = load_image(param[1].data.d_string);
    if (image_ID != -1) {
      *nreturn_vals = 2;
      values[0].data.d_status = STATUS_SUCCESS;
      values[1].type = PARAM_IMAGE;
      values[1].data.d_image = image_ID;
    } else {
      values[0].data.d_status = STATUS_EXECUTION_ERROR;
    }
  } else if (!strcmp(name, "file_lf2_save")) {
    if (!save_dialog()) {
      *nreturn_vals = 1;
      *return_vals = values;
      values[0].type = PARAM_STATUS;
      values[0].data.d_status = STATUS_EXECUTION_ERROR;
      return;
    }
    *nreturn_vals = 1;
    if (save_image(param[3].data.d_string,
		   param[1].data.d_int32, param[2].data.d_int32)) {
      values[0].data.d_status = STATUS_SUCCESS;
    } else {
      values[0].data.d_status = STATUS_EXECUTION_ERROR;
    }
  } else {
    g_assert (FALSE);
  }
}

static gint32 load_image (char *filename)
{
  LF2 *lf2;
  int i, j;
  int width, height;
  
  GPixelRgn pixel_rgn;
  GDrawable *drawable;
  gint32 image_ID;
  gint32 layer_ID;
  gchar *temp;
  guchar *gimp_cmap;
  FILE *fp;
  
  temp = g_malloc(strlen (filename) + 12);
  if (!temp) gimp_quit ();
  sprintf (temp, "Loading %s:", filename);
  gimp_progress_init (temp);
  g_free(temp);

  fp = fopen(filename, "r");
  if (fp == NULL) {
    return -1;
  }
  
  lf2 = read_lf2(fp);
  if (lf2 == NULL) {
    return -1;
  }
  gimp_cmap = g_malloc (sizeof (guchar) * 3 * LF2_COLOR_NUM);
  for (i = 0, j = 0; i < LF2_COLOR_NUM; i++) {
    gimp_cmap[j++] = lf2->palette[i][LF2_R];
    gimp_cmap[j++] = lf2->palette[i][LF2_G];
    gimp_cmap[j++] = lf2->palette[i][LF2_B];
  }
  
  width = lf2->width;
  height = lf2->height;
  
  image_ID = gimp_image_new (width, height, INDEXED);
  gimp_image_set_filename (image_ID, filename);
  if (lf2->transparent == 0xff) {
    layer_ID = gimp_layer_new (image_ID, "Background",
			       width,
			       height,
			       INDEXED_IMAGE, 100, NORMAL_MODE);
  } else {
    layer_ID = gimp_layer_new (image_ID, "Background",
			       width,
			       height,
			       INDEXEDA_IMAGE, 100, NORMAL_MODE);
  }
  gimp_image_add_layer (image_ID, layer_ID, 0);
  drawable = gimp_drawable_get (layer_ID);
  gimp_pixel_rgn_init (&pixel_rgn, drawable, 0, 0, drawable->width, drawable->height, TRUE, FALSE);
  gimp_image_set_cmap (image_ID, gimp_cmap, LF2_COLOR_NUM);

  lf2_expand_body(lf2, &pixel_rgn);

  gimp_drawable_flush (drawable);
  gimp_drawable_detach (drawable);

  free_lf2(lf2);
  
  return image_ID;
}

static LF2 *read_lf2(FILE *fp)
{
  LF2 *lf2;

  lf2 = init_lf2();
  
  if (read_magic(fp) < 0) {
    fclose(fp);
    return NULL;
  }
  read_size(fp, lf2);
  read_unknown(fp, lf2);

  skip_padding(fp, 1);
  read_transparent(fp, lf2);
  skip_padding(fp, 3);
  read_color_num(fp, lf2);
  skip_padding(fp, 1);
  read_palette(fp, lf2);

  if (extract_body(fp, lf2) < 0) {
    free_lf2(lf2);
    return NULL;
  }
  return lf2;
}

static LF2 *init_lf2(void)
{
  LF2 *lf2;

  lf2 = (LF2 *)malloc(sizeof(LF2));
  if (lf2 == NULL) {
    fprintf(stderr, "LF2: Error - Can't allocate memory.\n");
    return NULL;
  }
  return lf2;
}

static void free_lf2(LF2 *lf2)
{
  if (lf2 != NULL) {
    if (lf2->image != NULL) {
      free(lf2->image);
    }
    free(lf2);
  }
}

static int read_magic(FILE *fp)
{
  char buf[9];
  int i;

  for (i = 0; i < 8; i++) {
    buf[i] = fgetc(fp);
  }
  buf[8] = '\0';  /* for safe */
  
  if (strcmp("LEAF256", buf)) {
    fprintf(stderr, "This file isn't LF2 format.\n");
    return -1;
  }
  return 0;
}

static void read_unknown(FILE *fp, LF2 *lf2)
{
  lf2->unknown = fgetc(fp);
}

static void read_size(FILE *fp, LF2 *lf2)
{
  int upper, lower;

  /* x offset */
  upper = fgetc(fp);
  lower = fgetc(fp);
  lf2->xoffset = (lower << 8) | upper;

  /* y offset */
  upper = fgetc(fp);
  lower = fgetc(fp);
  lf2->yoffset = (lower << 8) | upper;

  /* width */
  upper = fgetc(fp);
  lower = fgetc(fp);
  lf2->width = (lower << 8) | upper;

  /* height */
  upper = fgetc(fp);
  lower = fgetc(fp);
  lf2->height = (lower << 8) | upper;

  /* set image size */
  lf2->size = lf2->width * lf2->height;
}

static void read_palette(FILE *fp, LF2 *lf2)
{
  int i;
  
  for (i = 0; i < lf2->color_num; i++) {
    lf2->palette[i][LF2_B] = fgetc(fp);
    lf2->palette[i][LF2_G] = fgetc(fp);
    lf2->palette[i][LF2_R] = fgetc(fp);
  }
}

static void read_color_num(FILE *fp, LF2 *lf2)
{
  lf2->color_num = fgetc(fp);
  if (lf2->color_num == 0x0) {
    lf2->color_num = 0x100;
  }
}

static void read_transparent(FILE *fp, LF2 *lf2)
{
  lf2->transparent = fgetc(fp);
}

static void skip_padding(FILE *fp, int num)
{
  int i;

  for (i = 0; i < num; i++) {
    fgetc(fp);
  }
}

static int extract_body(FILE *fp, LF2 *lf2)
{
  int ring[RINGBUF_SIZE];
  int i, j;
  int c, m;
  int flag;
  int upper, lower;
  int pos, len;
  int data;
  int idx;
  
  lf2->image = (int *)calloc(lf2->size, sizeof(int));
  if (lf2->image == NULL) {
    fprintf(stderr, "LF2: Error - Can't allocate memory.\n");
    return -1;
  }
  
  /* initialize ring buffer */
  for (i = 0; i < 0xfff; i++){
    ring[i] = 0;
  }
  
  /* extract data */
  for (i = 0, c = 0, m = 0xfee; i < lf2->size;){
    /* flag bits, which indicates data or location */
    if (--c < 0) {
      flag = fgetc(fp);
      flag ^= 0xff;
      c = 7;
    }
    
    if(flag & 0x80){
      /* data */
      data = fgetc(fp);
      if (data == EOF) {
	fprintf(stderr, "Data may be truncated.\n");
	return -1;
      }
      data ^= 0xff;
      ring[m++] = data;
      idx = (i % lf2->width) + lf2->width * (lf2->height - i / lf2->width - 1);
      lf2->image[idx] = data;
      i++;
      m &= 0xfff;
    } else {
      /* copy from ring buffer */
      upper = fgetc(fp);
      if (upper == EOF) {
	fprintf(stderr, "Data may be truncated.\n");
	return -1;
      }
      upper ^= 0xff;

      lower = fgetc(fp);
      if (lower == EOF) {
	fprintf(stderr, "Data may be truncated.\n");
	return -1;
      }
      lower ^= 0xff;
      
      len = (upper & 0x0f) + 3;
      pos = (upper >> 4) + (lower << 4);

      for(j = 0; j < len; j++){
	idx = (i % lf2->width) + lf2->width * (lf2->height - i / lf2->width - 1);
	ring[m++] = lf2->image[idx] =  ring[pos++];
	i++;
	m &= 0xfff;
	pos &= 0xfff;
      }
    }
    flag = flag << 1;
  }
  return 0;
}

static void lf2_expand_body(LF2 *lf2, GPixelRgn *pixel_rgn)
{
  int x, y;
  int width, height;
  int tile_height;
  guchar *dest, *dest_base;
  int row = 0;

  width = lf2->width;
  height = lf2->height;

  tile_height = gimp_tile_height();
  if (lf2->transparent == 0xff) {
    dest_base = dest = g_new(guchar, width * tile_height);
  } else {
    dest_base = dest = g_new(guchar, width * tile_height * 2);
  }

  for(y = 0; y < height; y++) {
    for(x = 0; x < width; x++) {
      *dest++ = lf2->image[y * width + x];
      if (lf2->transparent != 0xff) {
	*dest++ = (guchar)(lf2->image[y * width + x] == lf2->transparent ? 0 : 255);
      }
    }
    if (row >= tile_height - 1 || y >= height - 1) {
      gimp_pixel_rgn_set_rect (pixel_rgn, dest_base, 0, y - row, width, row + 1);
      gimp_progress_update ((double)y / (double)height);
      dest = dest_base;
      row = 0;
    } else {
      row++;
    }
  }
}

/*
 * save LF2 image
 */
gint save_image(char *filename, gint32 image_ID, gint32 drawable_ID)
{
  GPixelRgn pixel_rgn;
  GDrawable *drawable;
  GDrawableType drawable_type;
  FILE *outfile;
  guchar *cmap, *pixels, *pixmap;
  int numcols;
  char *temp_buf;
  LF2 *lf2;
  int i, idx;

  drawable = gimp_drawable_get (drawable_ID);
  drawable_type = gimp_drawable_type (drawable_ID);
  gimp_pixel_rgn_init (&pixel_rgn, drawable, 0, 0, drawable->width, drawable->height, FALSE, FALSE);
  cmap = gimp_image_get_cmap(image_ID, &numcols);
  
  temp_buf = g_malloc(strlen (filename) + 11);
  sprintf(temp_buf, "Saving %s:", filename);
  gimp_progress_init(temp_buf);
  g_free(temp_buf);

  lf2 = init_lf2();
  
  if (drawable_type != INDEXED_IMAGE && drawable_type != INDEXEDA_IMAGE) {
    fprintf(stderr, "LF2: Error - Only indexed image is supported.\n");
    return FALSE;
  }
  if (numcols > 256 && drawable_type == INDEXED_IMAGE) {
    fprintf(stderr, "LF2: Error - The palette size is limited to 256 colors.\n");
    return FALSE;
  } else if (numcols > 255 && drawable_type == INDEXEDA_IMAGE) {
    fprintf(stderr, "LF2: Error - The palette size is limited to 255 with a transparent color.\n");
    return FALSE;
  }

  outfile = fopen(filename, "wb");
  if (!outfile) {
    fprintf(stderr, "LF2: Error - Can't open %s.\n", filename);
    return FALSE;
  }

  lf2_set_cmap(lf2, numcols, cmap);
  
  /* get image body */
  if (drawable_type == INDEXEDA_IMAGE) {
    pixels = (guchar *)g_malloc(drawable->width * drawable->height * 2);
  } else {
    pixels = (guchar *)g_malloc(drawable->width * drawable->height);
  }
  gimp_pixel_rgn_get_rect(&pixel_rgn, pixels, 0, 0, drawable->width, drawable->height);
  
  lf2->width = drawable->width;
  lf2->height = drawable->height;
  lf2->size = lf2->width * lf2->height;
  lf2->color_num = numcols;
  if (drawable_type == INDEXEDA_IMAGE) {
    lf2->transparent = numcols;
  } else {
    lf2->transparent = 0xff;
  }
  lf2->xoffset = lf2svals.xoffset;
  lf2->yoffset = lf2svals.yoffset;
  lf2->encoded_data = (int *)calloc(lf2->size, sizeof(int));

  pixmap = (guchar *)g_malloc(drawable->width * drawable->height);
  if (drawable_type == INDEXEDA_IMAGE) {
    for (i = 0; i < drawable->width * drawable->height; i++) {
      idx = (i % drawable->width) + drawable->width * (drawable->height - i / drawable->width - 1);
      if (pixels[i * 2 + 1] == 0) { /* transparent */
	pixmap[idx] = lf2->transparent;
      } else {
	pixmap[idx] = pixels[i * 2];
      }
    }
  } else {
    for (i = 0; i < drawable->width * drawable->height; i++) {
      idx = (i % drawable->width) + drawable->width * (drawable->height - i / drawable->width - 1);
      pixmap[idx] = pixels[i];
    }
  }
  free(pixels);

  lf2_encode(pixmap, lf2);
  free(pixmap);

  output_lf2(lf2, outfile);
  gimp_progress_update (1.0);  

  free(lf2->encoded_data);
  free_lf2(lf2);
  return TRUE;
}

static void lf2_set_cmap(LF2 *lf2, int numcols, guchar *cmap)
{
  int i;
  for (i = 0; i < numcols; i++) {
    lf2->palette[i][LF2_R] = *cmap++;
    lf2->palette[i][LF2_G] = *cmap++;
    lf2->palette[i][LF2_B] = *cmap++;
  }
}

static int output_lf2(LF2 *lf2, FILE *fp)
{
  int i;

  fprintf(fp, "LEAF256");
  fputc('\0', fp);
  output_word(lf2->xoffset, fp);
  output_word(lf2->yoffset, fp);
  output_word(lf2->width, fp);
  output_word(lf2->height, fp);
  fputc(0xa8, fp);
  fputc(0x00, fp);
  fputc(lf2->transparent, fp);
  fputc(0x00, fp);
  fputc(0x00, fp);
  fputc(0x00, fp);
  fputc(lf2->color_num, fp);
  fputc(0x00, fp);
  
  for (i = 0; i < lf2->color_num; i++) {
    fputc(lf2->palette[i][LF2_B], fp);
    fputc(lf2->palette[i][LF2_G], fp);
    fputc(lf2->palette[i][LF2_R], fp);
  }

  for (i = 0; i < lf2->encoded_size; i++) {
    fputc(lf2->encoded_data[i] ^ 0xff, fp);
  }

  return 0;
}

static void output_word(int d, FILE *fp)
{
  int d1, d2;

  d1 = d & 0xff;
  d2 = (d >> 8) & 0xff;

  fputc(d1, fp);
  fputc(d2, fp);
}

/*
 * initialize a tree
 */
void init_tree(void)
{
  int i;
  
  for (i = RINGBUF_SIZE + 1; i <= RINGBUF_SIZE + 256; i++) {
    rson[i] = NIL;
  }
  for (i = 0; i < RINGBUF_SIZE; i++) {
    dad[i] = NIL;
  }
}

/*
 * insert a node r into a tree
 */
void insert_node(int r)
{
  int i, p, cmp;
  unsigned char *key;
  
  cmp = 1;
  key = &text[r];
  p = RINGBUF_SIZE + 1 + key[0];
  rson[r] = lson[r] = NIL;
  matchlen = 0;

  for ( ; ; ) {
    if (cmp >= 0) {
      if (rson[p] != NIL) p = rson[p];
      else {  rson[p] = r;  dad[r] = p;  return;  }
    } else {
      if (lson[p] != NIL) p = lson[p];
      else {  lson[p] = r;  dad[r] = p;  return;  }
    }
    for (i = 1; i < F; i++)
      if ((cmp = key[i] - text[p + i]) != 0)  break;
    if (i > matchlen) {
      matchpos = p;
      if ((matchlen = i) >= F)  break;
    }
  }
  dad[r] = dad[p];  lson[r] = lson[p];  rson[r] = rson[p];
  dad[lson[p]] = r;  dad[rson[p]] = r;
  if (rson[dad[p]] == p) {
    rson[dad[p]] = r;
  } else {
    lson[dad[p]] = r;
  }
  /* detach p */
  dad[p] = NIL;
}

/*
 * delete a node p from tree
 */
void delete_node(int p)
{
  int  q;
  
  if (dad[p] == NIL) { /* p isn't found */
    return;
  }
  
  if (rson[p] == NIL) {
    q = lson[p];
  } else if (lson[p] == NIL) {
    q = rson[p];
  } else {
    q = lson[p];
    if (rson[q] != NIL) {
      do {
	q = rson[q];
      } while (rson[q] != NIL);
      rson[dad[q]] = lson[q];
      dad[lson[q]] = dad[q];
      lson[q] = lson[p];
      dad[lson[p]] = q;
    }
    rson[q] = rson[p];
    dad[rson[p]] = q;
  }
  dad[q] = dad[p];
  if (rson[dad[p]] == p) {
    rson[dad[p]] = q;
  } else {
    lson[dad[p]] = q;
  }
  dad[p] = NIL;
}

/*
 * main routine of LF2 encoder
 */
void lf2_encode(guchar *img, LF2 *lf2)
{
  int i, c, len, r, s, lastmatchlen, codeptr;
  unsigned char code[17], mask;
  unsigned long int incount = 0;
  int img_ptr = 0;
  int lf2_ptr = 0;

  /* initialize a tree */
  init_tree();
  
  code[0] = 0;
  codeptr = 1;
  mask = 0x80;
  s = 0;
  r = RINGBUF_SIZE - F;
  
  /* initialize a ring buffer */
  for (i = s; i < r; i++) {
    text[i] = 0x0;
  }
  for (len = 0; len < F ; len++) {
    c = img[img_ptr++];
    if (img_ptr > lf2->size) break;
    text[r + len] = c;
  }

  incount = len;
  if (incount == 0) {
    return;
  }
  for (i = 1; i <= F; i++) insert_node(r - i);
  insert_node(r);
  
  do {
    if (matchlen > len) {
      matchlen = len;
    }
    if (matchlen < 3) {
      matchlen = 1;
      code[0] |= mask;
      code[codeptr++] = text[r];
    } else {
      code[codeptr++] = (unsigned char)(((matchpos & 0x0f) << 4)| (matchlen - 3));
      code[codeptr++] = (unsigned char)(matchpos >> 4);
    }
    if ((mask >>= 1) == 0) {
      for (i = 0; i < codeptr; i++) {
	lf2->encoded_data[lf2_ptr++] = code[i];
      }
      outcount += codeptr;
      code[0] = 0;
      codeptr = 1;
      mask = 0x80;
    }
    lastmatchlen = matchlen;
    for (i = 0; i < lastmatchlen; i++) {
      c = img[img_ptr++];
      if (img_ptr > lf2->size) {
	break;
      }
      delete_node(s);
      text[s] = c;
      if (s < F - 1) {
	text[s + RINGBUF_SIZE] = c;
      }
      s = (s + 1) & (RINGBUF_SIZE - 1);
      r = (r + 1) & (RINGBUF_SIZE - 1);
      insert_node(r);
    }
    if (img_ptr % 1000 == 0) {
      gimp_progress_update ((double) img_ptr / (double) lf2->size);
    }
    while (i++ < lastmatchlen) {
      delete_node(s);
      s = (s + 1) & (RINGBUF_SIZE - 1);
      r = (r + 1) & (RINGBUF_SIZE - 1);
      if (--len) insert_node(r);
    }
  } while (len > 0);
  
  if (codeptr > 1) {
    for (i = 0; i < codeptr; i++) {
      lf2->encoded_data[lf2_ptr++] = code[i];
    }
    outcount += codeptr;
  }

  lf2->encoded_size = lf2_ptr;
}

/*
 * save dialog
 */

static gint save_dialog(void)
{
  gchar **argv;
  gint argc;
  gchar buffer[LF2_BUFLEN];

  GtkWidget *dlg, *button, *label, *entry;
  GtkWidget *frame, *frame_vbox, *hbox;
  
  argc = 1;
  argv = g_new(gchar *, 1);
  argv[0] = g_strdup("lf2");

  gtk_init(&argc, &argv);
  
  dlg = gtk_dialog_new();
  gtk_window_set_title(GTK_WINDOW (dlg), "Save LF2 file");
  gtk_window_position(GTK_WINDOW (dlg), GTK_WIN_POS_MOUSE);
  gtk_signal_connect(GTK_OBJECT (dlg), "destroy",
		      (GtkSignalFunc)save_close_callback, NULL);
  /*
   * action area 
   */
  button = gtk_button_new_with_label ("OK");
  GTK_WIDGET_SET_FLAGS(button, GTK_CAN_DEFAULT);
  gtk_signal_connect(GTK_OBJECT (button), "clicked",
		     (GtkSignalFunc)save_ok_callback,
		     dlg);
  gtk_box_pack_start(GTK_BOX (GTK_DIALOG (dlg)->action_area), button, TRUE, TRUE, 0);
  gtk_widget_grab_default (button);
  gtk_widget_show(button);
  
  button = gtk_button_new_with_label("Cancel");
  GTK_WIDGET_SET_FLAGS (button, GTK_CAN_DEFAULT);
  gtk_signal_connect_object(GTK_OBJECT (button), "clicked",
			    (GtkSignalFunc) gtk_widget_destroy,
			    GTK_OBJECT (dlg));
  gtk_box_pack_start(GTK_BOX (GTK_DIALOG (dlg)->action_area), button, TRUE, TRUE, 0);
  gtk_widget_show (button);

  /*
   * parameter settings
   */
  frame = gtk_frame_new("Offset");
  gtk_frame_set_shadow_type(GTK_FRAME(frame), GTK_SHADOW_ETCHED_IN);
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG (dlg)->vbox), frame, TRUE, TRUE, 0);

  frame_vbox = gtk_vbox_new(FALSE, 5);
  gtk_container_add(GTK_CONTAINER(frame), frame_vbox);
  
  /*
   * create a series of buttons and labels
   */
  hbox = gtk_hbox_new(FALSE, 5);
  gtk_container_add(GTK_CONTAINER(frame_vbox), hbox);

  /* label 1 */
  label = gtk_label_new("x:");
  gtk_box_pack_start(GTK_BOX(hbox), label, TRUE, TRUE, 5);
  gtk_widget_show(label);

  /* text entry 1 */
  entry = gtk_entry_new ();
  gtk_box_pack_start (GTK_BOX(hbox), entry, TRUE, TRUE, 5);
  gtk_widget_set_usize (entry, LF2_ENTRY_WIDTH, 0);
  sprintf (buffer, "%d", lf2svals.xoffset);
  gtk_entry_set_text (GTK_ENTRY(entry), buffer);
  gtk_signal_connect (GTK_OBJECT(entry), "changed",
                      (GtkSignalFunc) lf2_xentry_callback,
                      NULL);
  gtk_widget_show (entry);

  gtk_widget_show(hbox);

  hbox = gtk_hbox_new(FALSE, 5);
  gtk_container_add(GTK_CONTAINER(frame_vbox), hbox);

  /* label 2 */
  label = gtk_label_new("y:");
  gtk_box_pack_start(GTK_BOX(hbox), label, TRUE, TRUE, 5);
  gtk_widget_show(label);
  
  /* button 2 */
  entry = gtk_entry_new ();
  gtk_box_pack_start (GTK_BOX(hbox), entry, TRUE, TRUE, 5);
  gtk_widget_set_usize (entry, LF2_ENTRY_WIDTH, 0);
  sprintf (buffer, "%d", lf2svals.yoffset);
  gtk_entry_set_text (GTK_ENTRY(entry), buffer);
  gtk_signal_connect (GTK_OBJECT(entry), "changed",
                      (GtkSignalFunc) lf2_yentry_callback,
                      NULL);
  gtk_widget_show (entry);

  gtk_widget_show(hbox);
  gtk_widget_show(frame_vbox);
  gtk_widget_show(frame);
  gtk_widget_show (dlg);

  gtk_main ();
  gdk_flush ();
  
  return lf2sint.run;
}

static void save_close_callback (GtkWidget *widget, gpointer data)
{
  gtk_main_quit ();
}

static void save_ok_callback (GtkWidget *widget, gpointer data)
{
  lf2sint.run = TRUE;
  gtk_widget_destroy (GTK_WIDGET (data));
}

static void lf2_xentry_callback (GtkWidget *widget, gpointer data)
{
  lf2svals.xoffset = atoi(gtk_entry_get_text(GTK_ENTRY(widget)));
}

static void lf2_yentry_callback (GtkWidget *widget, gpointer data)
{
  lf2svals.yoffset = atoi(gtk_entry_get_text(GTK_ENTRY(widget)));
}
