/*
 * lfg.c, LEAF LFG plug-in for GIMP
 * Copyright (C) 1997 TF <tf@denpa.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * lfg.c supports the LFG format images (INDEX/INDEXA). This format
 * is used in the LEAF's games(SHIZUKU, KIZUATO.)
 *
 * Changes:
 *   11/15/1997 Support output
 *   09/25/1997 Original code is written
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <setjmp.h>
#include "gtk/gtk.h"
#include "libgimp/gimp.h"


#define R 0
#define G 1
#define B 2

#define VERTICAL 0
#define HORIZONTAL 1

#define PADDING_LENGTH 2
#define LFG_COLOR_NUM 16

#define LFG_ENTRY_WIDTH 64
#define LFG_BUFLEN 64

#define RINGBUF_SIZE 0x1000    /* size of ring buffer */
#define F 18    /* longest match length */
#define NIL RINGBUF_SIZE    /* end of tree */

typedef struct {
  int width;
  int height;
  int xoffset;
  int yoffset;
  int direction;       /* vertical or horizontal */
  int transparent;     /* the palette num. of transparent color */
  int size;            /* the size of image */
  int palette[LFG_COLOR_NUM][3];
  int *body;           /* extracted data */
  int *image;          /* indexed image */
  int *encoded_data;
  int encoded_size;
} LFG;

/*
 * plug-in interfaces
 */
static void query(void);
static void run(char *, int, GParam *, int *, GParam **);
static gint32 load_image (char *);
gint save_image(char *, gint32, gint32);

/*
 * dialog and callbacks
 */
static gint save_dialog(void);
static void save_ok_callback(GtkWidget *, gpointer);
static void save_close_callback(GtkWidget *, gpointer);
static void lfg_xentry_callback (GtkWidget *, gpointer);
static void lfg_yentry_callback (GtkWidget *, gpointer);
static void toggle_update(GtkWidget *, gpointer);

/*
 * read LFG
 */
static LFG *read_lfg(FILE *);
static void free_lfg(LFG *);
static void read_magic(FILE *);
static void read_palette(FILE *, LFG *);
static void read_size(FILE *, LFG *);
static void read_type(FILE *, LFG *);
static void read_transparent(FILE *, LFG *);
static void skip_padding(FILE *, int);
static void read_image_size(FILE *, LFG *);
static void extract_body(FILE *, LFG *);
static void convert_indexed_image(LFG *);
static void lfg_expand_body(LFG *, GPixelRgn *);

/*
 * output LFG
 */
static int output_lfg(LFG *, FILE *);
void init_tree(void);
void insert_node(int);
void delete_node(int);
static LFG* init_lfg(void);
static void lfg_encode(guchar *, LFG *);
static void lfg_set_cmap(LFG *, int, guchar *);
static guchar optimal_color_component(guchar);
static guchar *pack_pixel(guchar *, int, int, int);
static guchar pack(guchar, guchar);
static void output_word(int, FILE *);
static void output_longword(int, FILE *);
/*
 * plug-in interfaces
 */
GPlugInInfo PLUG_IN_INFO = {
  NULL,    /* init_proc */
  NULL,    /* quit_proc */
  query,   /* query_proc */
  run,     /* run_proc */
};

typedef struct {
  gint run;
} LFGSaveInterface;

static LFGSaveInterface lfgsint = {
  FALSE   /*  run  */
};

typedef struct {
  int xoffset;
  int yoffset;
} LFGSaveVals;

static LFGSaveVals lfgsvals = {
  0,    /* xoffset */
  0,    /* yoffset */
};

static int tgl_idx[] = {TRUE, FALSE};

/*
 * for encoding
 */
unsigned long outcount = 0;  /* counter for output size */
unsigned char text[RINGBUF_SIZE+F-1]; /* text buffer */
int dad[RINGBUF_SIZE+1];
int lson[RINGBUF_SIZE+1];
int rson[RINGBUF_SIZE+257];  /* tree */

int matchpos;  /* the position of longest match pattern */
int matchlen;  /* the length of longest match pattern */

MAIN();

static void query(void)
{
  static GParamDef load_args[] = {
    { PARAM_INT32, "run_mode", "Interactive, non-interactive" },
    { PARAM_STRING, "filename", "The name of the file to load" },
    { PARAM_STRING, "raw_filename", "The name entered" },
  };

  static GParamDef load_return_vals[] = {
    { PARAM_IMAGE, "image", "Output image" },
  };

  static GParamDef save_args[] = {
    { PARAM_INT32, "run_mode", "Interactive, non-interactive" },
    { PARAM_IMAGE, "image", "Input image" },
    { PARAM_DRAWABLE, "drawable", "Drawable to save" },
    { PARAM_STRING, "filename", "The name of the file to save the image in" },
    { PARAM_STRING, "raw_filename", "The name entered" },
  };

  static int nload_args = sizeof (load_args) / sizeof (load_args[0]);
  static int nload_return_vals = sizeof (load_return_vals) / sizeof (load_return_vals[0]);
  static int nsave_args = sizeof (save_args) / sizeof (save_args[0]);

  gimp_install_procedure ("file_lfg_load",
                          "loads files of the lfg file format",
                          "FIXME: write help for lfg_load",
                          "FUJIWARA Teruyoshi",
                          "FUJIWARA Teruyoshi",
                          "1997",
                          "<Load>/LFG",
			  NULL,
                          PROC_PLUG_IN,
                          nload_args, nload_return_vals,
                          load_args, load_return_vals);

  gimp_install_procedure ("file_lfg_save",
                          "saves files in lfg file format",
                          "FIXME: write help for lfg_save",
                          "FUJIWARA Teruyoshi",
                          "FUJIWARA Teruyoshi",
                          "1997",
                          "<Save>/LFG",
			  "INDEXED*",
                          PROC_PLUG_IN,
                          nsave_args, 0,
                          save_args, NULL);

  gimp_register_magic_load_handler ("file_lfg_load", "lfg", "", "0,string,LEAFCODE");
  gimp_register_save_handler ("file_lfg_save", "lfg", "");

}

static void run (char *name, int nparams, GParam *param, int *nreturn_vals, GParam **return_vals)
{
  static GParam values[2];
  GRunModeType run_mode;
  gint32 image_ID;
  
  run_mode = param[0].data.d_int32;
  
  *nreturn_vals = 1;
  *return_vals = values;
  
  values[0].type = PARAM_STATUS;
  values[0].data.d_status = STATUS_CALLING_ERROR;
  
  if (!strcmp(name, "file_lfg_load")) {
    image_ID = load_image(param[1].data.d_string);
    if (image_ID != -1) {
      *nreturn_vals = 2;
      values[0].data.d_status = STATUS_SUCCESS;
      values[1].type = PARAM_IMAGE;
      values[1].data.d_image = image_ID;
    } else {
      values[0].data.d_status = STATUS_EXECUTION_ERROR;
    }
  } else if (!strcmp(name, "file_lfg_save")) {
    if (!save_dialog()) {
      *nreturn_vals = 1;
      *return_vals = values;
      values[0].type = PARAM_STATUS;
      values[0].data.d_status = STATUS_EXECUTION_ERROR;
      return;
    }
    *nreturn_vals = 1;
    if (save_image(param[3].data.d_string,
		   param[1].data.d_int32, param[2].data.d_int32)) {
      values[0].data.d_status = STATUS_SUCCESS;
    } else {
      values[0].data.d_status = STATUS_EXECUTION_ERROR;
    }
  } else {
    g_assert (FALSE);
  }
}

static gint32 load_image (char *filename)
{
  LFG *lfg;
  int i, j;
  int width, height;
  
  GPixelRgn pixel_rgn;
  GDrawable *drawable;
  gint32 image_ID;
  gint32 layer_ID;
  gchar *temp;
  guchar *gimp_cmap;
  FILE *fp;
  
  temp = g_malloc(strlen (filename) + 12);
  if (!temp) gimp_quit ();
  sprintf (temp, "Loading %s:", filename);
  gimp_progress_init (temp);
  g_free(temp);

  fp = fopen(filename, "r");
  if (fp == NULL) {
    return 0;
  }
  
  lfg = read_lfg(fp);
  gimp_cmap = g_malloc (sizeof (guchar) * 3 * LFG_COLOR_NUM);
  for (i = 0, j = 0; i < LFG_COLOR_NUM; i++) {
    gimp_cmap[j++] = lfg->palette[i][R];
    gimp_cmap[j++] = lfg->palette[i][G];
    gimp_cmap[j++] = lfg->palette[i][B];
  }
  
  width = lfg->width - lfg->xoffset;
  height = lfg->height - lfg->yoffset;
  
  image_ID = gimp_image_new (width, height, INDEXED);
  gimp_image_set_filename (image_ID, filename);
  if (lfg->transparent == 0xff) {
    layer_ID = gimp_layer_new (image_ID, "Background",
			       width,
			       height,
			       INDEXED_IMAGE, 100, NORMAL_MODE);
  } else {
    layer_ID = gimp_layer_new (image_ID, "Background",
			       width,
			       height,
			       INDEXEDA_IMAGE, 100, NORMAL_MODE);
  }
  gimp_image_add_layer (image_ID, layer_ID, 0);
  drawable = gimp_drawable_get (layer_ID);
  gimp_pixel_rgn_init (&pixel_rgn, drawable, 0, 0, drawable->width, drawable->height, TRUE, FALSE);
  gimp_image_set_cmap (image_ID, gimp_cmap, LFG_COLOR_NUM);

  lfg_expand_body(lfg, &pixel_rgn);
  
  gimp_drawable_flush (drawable);
  gimp_drawable_detach (drawable);

  free_lfg(lfg);
  
  return image_ID;
}

LFG *read_lfg(FILE *fp)
{
  LFG *lfg;

  lfg = (LFG *)malloc(sizeof(LFG));
  if (lfg == NULL) {
    fprintf(stderr, "LFG: Error - Can't allocate memory.\n");
    exit(1);
  }

  read_magic(fp);
  read_palette(fp, lfg);
  read_size(fp, lfg);
  read_type(fp, lfg);
  read_transparent(fp, lfg);
  skip_padding(fp, PADDING_LENGTH);
  read_image_size(fp, lfg);

  extract_body(fp, lfg);
  
  convert_indexed_image(lfg);

  return lfg;
}

void free_lfg(LFG *lfg)
{
  if (lfg == NULL) {
    fprintf(stderr, "Warning: NULL LFG file.\n");
    return;
  }

  if (lfg->body != NULL) { /* Normally, lfg->body should be NULL. */
    free(lfg->image);
  }
  
  if (lfg->image != NULL) {
    free(lfg->image);
  }

  free(lfg);
}

/*
 * Check if this is LFG file.
 */
static void read_magic(FILE *fp)
{
  char buf[9];
  int i;

  for (i = 0; i < 8; i++) {
    buf[i] = fgetc(fp);
  }
  buf[8] = '\0';
  
  if (strcmp("LEAFCODE", buf)) {
    fprintf(stderr, "This file isn't LFG format.\n");
    exit(1);
  }
}

static void read_palette(FILE *fp, LFG *lfg)
{
  int i, j = 0, rgb = R;
  int c, upper, lower;
  
  for (i = 0; i < 24; i++) {
    c = fgetc(fp);
    upper = (c & 0xf0) >> 4;
    lower = c & 0xf;

    /* upper byte */
    lfg->palette[j][rgb] = (upper << 4) | upper;
    rgb = (rgb + 1) % 3;
    if (rgb == 0) {
      j++;
    }

    /* lower byte */
    lfg->palette[j][rgb] = (lower << 4) | lower;
    rgb = (rgb + 1) % 3;
    if (rgb == 0) {
      j++;
    }
  }
}

static void read_size(FILE *fp, LFG *lfg)
{
  int upper, lower;

  /* x offset */
  upper = fgetc(fp);
  lower = fgetc(fp);
  lfg->xoffset = (lower << 8 | upper) * 8;

  /* y offset */
  upper = fgetc(fp);
  lower = fgetc(fp);
  lfg->yoffset = lower << 8 | upper;

  /* width */
  upper = fgetc(fp);
  lower = fgetc(fp);
  lfg->width = ((lower << 8 | upper) + 1) * 8;

  /* height */
  upper = fgetc(fp);
  lower = fgetc(fp);
  lfg->height = (lower << 8 | upper) + 1;

}

/*
 * the direction pixels are tiled has two types,
 * vertical(per 2 pixels) or horizontal.
 */
static void read_type(FILE *fp, LFG *lfg)
{
  if (fgetc(fp) == 0) {
    lfg->direction = VERTICAL;
  } else {
    lfg->direction = HORIZONTAL;
  }
}

static void read_transparent(FILE *fp, LFG *lfg)
{
  lfg->transparent = fgetc(fp);
  if (lfg->transparent == 0xff) {
    /* Transparent color isn't used */
  } else if (lfg->transparent < 0 || lfg->transparent > 15) {
    fprintf(stderr, "palette number of transparent color is invalid.\n");
    exit(1);
  }
}

static void skip_padding(FILE *fp, int length)
{
  int i;

  for (i = 0; i < length; i++) {
    fgetc(fp);
  }
  return;
}

/*
 * Read the data size after extraction.
 */
static void read_image_size(FILE *fp, LFG *lfg)
{
  int a, b, c, d;
  
  a = fgetc(fp);
  b = fgetc(fp);
  c = fgetc(fp);
  d = fgetc(fp);
  
  lfg->size = a | (b << 8) | (c << 16) | (d << 24);
}

static void extract_body(FILE *fp, LFG *lfg)
{
  int ring[0x1000];
  int i, j;
  int c, m;
  int flag;
  int upper, lower;
  int pos, len;
  
  lfg->body = (int *)calloc(lfg->size, sizeof(int));
  if (lfg->body == NULL) {
    fprintf(stderr, "extract_body: Can't allocate memory.\n");
    exit(1);
  }
  
  /* initialize ring buffer */
  for (i = 0; i < 0xfff; i++){
    ring[i] = 0;
  }
  
  /* extract data */
  for (i = 0, c = 0, m = 0xfee; i < lfg->size;){
    /* flag bits, which indicates data or location */
    if (--c < 0){
      flag = fgetc(fp);
      c = 7;
    }
    
    if(flag & 0x80){
      /* data */
      ring[m++] = lfg->body[i++] = fgetc(fp);
      m &= 0xfff;
    } else {
      /* copy from ring buffer */
      upper = fgetc(fp);
      lower = fgetc(fp);
      
      len = (upper & 0x0f) + 3;
      pos = (upper >> 4) + (lower << 4);
      if (pos < 0 || pos > 0xfff) { /* must always be false. */
	fprintf(stderr, "Invalid position.\n");
	exit(1);
      }
      
      for(j = 0; j < len; j++){
	ring[m++] = lfg->body[i++] = ring[pos++];
	m &= 0xfff;
	pos &= 0xfff;
      }
    }
    flag = flag << 1;
  }

  /* check if the extracted size is correct. */
  if (i != lfg->size) {
    fprintf(stderr, "Warning: Extracted data may be invalid.\n");
  }
}

static void convert_indexed_image(LFG *lfg)
{
  int i;
  int pix1, pix2;
  int x = 0, y = 0;

  lfg->image = (int *)calloc((lfg->width - lfg->xoffset) * (lfg->height - lfg->yoffset), sizeof(int));
  if (lfg->image == NULL) {
    fprintf(stderr, "convert_indexed_image: Can't allocate memory.\n");
    exit(1);
  }
  
  for (i = 0; i < lfg->size; i++) {
    pix1 = (lfg->body[i] & 0x80) >> 4 |
      (lfg->body[i] & 0x20) >> 3 |
      (lfg->body[i] & 0x08) >> 2 |
      (lfg->body[i] & 0x02) >> 1;
    pix2 = (lfg->body[i] & 0x40) >> 3 |
      (lfg->body[i] & 0x10) >> 2 |
      (lfg->body[i] & 0x04) >> 1 |
      (lfg->body[i] & 0x01);
    if (lfg->direction == VERTICAL) {
      lfg->image[(lfg->width - lfg->xoffset) * y + x] = pix1;
      lfg->image[(lfg->width - lfg->xoffset) * y + x + 1] = pix2;
      if (++y > lfg->height - lfg->yoffset - 1) {
	y = 0;
	x += 2;
      }
    } else { /* HORIZONTAL */
      lfg->image[(lfg->width - lfg->xoffset) * y + x] = pix1;
      lfg->image[(lfg->width - lfg->xoffset) * y + x + 1] = pix2;
      x += 2;
      if (x > (lfg->width - lfg->xoffset) - 1) {
	x = 0;
	y++;
      }
    }
  }

  free(lfg->body);
  lfg->body = NULL;
}

static void lfg_expand_body(LFG *lfg, GPixelRgn *pixel_rgn)
{
  int x, y;
  int width, height;
  int tile_height;
  guchar *dest, *dest_base;
  int row = 0;

  width = lfg->width - lfg->xoffset;
  height = lfg->height - lfg->yoffset;

  tile_height = gimp_tile_height();
  if (lfg->transparent == 0xff) {
    dest_base = dest = g_new(guchar, width * tile_height);
  } else {
    dest_base = dest = g_new(guchar, width * tile_height * 2);
  }

  for(y = 0; y < height; y++) {
    for(x = 0; x < width; x++) {
      *dest++ = lfg->image[y * width + x];
      if (lfg->transparent != 0xff) {
	*dest++ = (guchar)(lfg->image[y * width + x] == lfg->transparent ? 0 : 255);
      }
    }
    if (row >= tile_height - 1 || y >= height - 1) {
      gimp_pixel_rgn_set_rect (pixel_rgn, dest_base, 0, y - row, width, row + 1);
      gimp_progress_update ((double) y / (double) height);
      dest = dest_base;
      row = 0;
    } else {
      row++;
    }
  }
}

/*
 * save LFG image
 */
gint save_image(char *filename, gint32 image_ID, gint32 drawable_ID)
{
  GPixelRgn pixel_rgn;
  GDrawable *drawable;
  GDrawableType drawable_type;
  FILE *outfile;
  guchar *cmap, *pixels, *pixmap, *packed_pixmap;
  int numcols;
  char *temp_buf;
  LFG *lfg;
  int i;

  drawable = gimp_drawable_get (drawable_ID);
  drawable_type = gimp_drawable_type (drawable_ID);
  gimp_pixel_rgn_init(&pixel_rgn, drawable, 0, 0, drawable->width, drawable->height, FALSE, FALSE);
  cmap = gimp_image_get_cmap(image_ID, &numcols);
  
  temp_buf = g_malloc(strlen (filename) + 11);
  sprintf(temp_buf, "Saving %s:", filename);
  gimp_progress_init(temp_buf);
  g_free(temp_buf);

  if (drawable_type != INDEXED_IMAGE && drawable_type != INDEXEDA_IMAGE) {
    fprintf(stderr, "LFG: Error - Only indexed image is supported.\n");
    return FALSE;
  }
  if (drawable_type == INDEXED_IMAGE && numcols > 16) {
    fprintf(stderr, "LFG: Error - This format is limited to 16 colors.\n");
    return FALSE;
  } else if (drawable_type == INDEXEDA_IMAGE && numcols > 15) {
    fprintf(stderr, "LFG: Error - With a transperent color, the maximum palette size is 15.\n");
    return FALSE;
  }

  /*  Should this be an error? */
  if (drawable->width % 8 != 0) {
    fprintf(stderr, "LFG: Error - Image width(%d) should be a multiple of 8.\n", drawable->width);
    return FALSE;
  }

  outfile = fopen(filename, "wb");
  if (!outfile) {
    fprintf(stderr, "LFG: Error - Can't open %s.\n", filename);
    return FALSE;
  }

  lfg = init_lfg();

  lfg_set_cmap(lfg, numcols, cmap);
  
  /* get image body */
  if (drawable_type == INDEXEDA_IMAGE) {
    pixels = (guchar *)g_malloc(drawable->width * drawable->height * 2);
  } else {
    pixels = (guchar *)g_malloc(drawable->width * drawable->height);
  }
  gimp_pixel_rgn_get_rect(&pixel_rgn, pixels, 0, 0, drawable->width, drawable->height);

  /* image size */
  lfg->xoffset = lfgsvals.xoffset;
  lfg->yoffset = lfgsvals.yoffset;
  lfg->width = drawable->width + lfg->xoffset;
  lfg->height = drawable->height + lfg->yoffset;
  lfg->size = drawable->width * drawable->height / 2;

  /* transparent color */
  if (drawable_type == INDEXEDA_IMAGE) {
    lfg->transparent = numcols;
  } else {
    lfg->transparent = 0xff;
  }

  /* direction */
  if (tgl_idx[0]) {
    lfg->direction = VERTICAL;
  }
  if (tgl_idx[1]) {
    lfg->direction = HORIZONTAL;
  }

  /*
   * create indexed image including transparent color
   */
  if (drawable_type == INDEXEDA_IMAGE) {
    pixmap = (guchar *)g_malloc(drawable->width * drawable->height);
    for (i = 0; i < drawable->width * drawable->height; i++) {
      if (pixels[i * 2 + 1] == 0) { /* transparent */
	pixmap[i] = lfg->transparent;
      } else {
	pixmap[i] = pixels[i * 2];
      }
    }
    free(pixels);
  } else {
    pixmap = pixels;
  }
  
  packed_pixmap = pack_pixel(pixmap, drawable->width, drawable->height, lfg->direction);
  free(pixmap);

  lfg->encoded_data = (int *)calloc(lfg->size, sizeof(int));
  lfg_encode(packed_pixmap, lfg);
  free(packed_pixmap);

  output_lfg(lfg, outfile);
  gimp_progress_update (1.0);  
  
  free(lfg->encoded_data);
  free_lfg(lfg);

  return TRUE;
}

/*
 * save dialog
 */
static gint save_dialog(void)
{
  gchar **argv;
  gint argc;
  gchar buffer[LFG_BUFLEN];

  GtkWidget *dlg, *button, *label, *entry;
  GtkWidget *frame, *frame_vbox, *hbox;
  GSList *group = NULL;

  argc = 1;
  argv = g_new(gchar *, 1);
  argv[0] = g_strdup("lfg");

  gtk_init(&argc, &argv);
  
  dlg = gtk_dialog_new();
  gtk_window_set_title(GTK_WINDOW (dlg), "Save LFG file");
  gtk_window_position(GTK_WINDOW (dlg), GTK_WIN_POS_MOUSE);
  gtk_signal_connect(GTK_OBJECT (dlg), "destroy",
		      (GtkSignalFunc)save_close_callback, NULL);
  /*
   * action area 
   */
  button = gtk_button_new_with_label ("OK");
  GTK_WIDGET_SET_FLAGS(button, GTK_CAN_DEFAULT);
  gtk_signal_connect(GTK_OBJECT (button), "clicked",
		     (GtkSignalFunc)save_ok_callback,
		     dlg);
  gtk_box_pack_start(GTK_BOX (GTK_DIALOG (dlg)->action_area), button, TRUE, TRUE, 0);
  gtk_widget_grab_default (button);
  gtk_widget_show(button);
  
  button = gtk_button_new_with_label("Cancel");
  GTK_WIDGET_SET_FLAGS (button, GTK_CAN_DEFAULT);
  gtk_signal_connect_object(GTK_OBJECT (button), "clicked",
			    (GtkSignalFunc) gtk_widget_destroy,
			    GTK_OBJECT (dlg));
  gtk_box_pack_start(GTK_BOX (GTK_DIALOG (dlg)->action_area), button, TRUE, TRUE, 0);
  gtk_widget_show (button);

  /*
   * parameter settings
   */
  frame = gtk_frame_new("Offset");
  gtk_frame_set_shadow_type(GTK_FRAME(frame), GTK_SHADOW_ETCHED_IN);
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG (dlg)->vbox), frame, TRUE, TRUE, 0);

  frame_vbox = gtk_vbox_new(FALSE, 5);
  gtk_container_add(GTK_CONTAINER(frame), frame_vbox);
  
  /*
   * create a series of buttons and labels
   */
  hbox = gtk_hbox_new(FALSE, 5);
  gtk_container_add(GTK_CONTAINER(frame_vbox), hbox);

  /* label 1 */
  label = gtk_label_new("x:");
  gtk_box_pack_start(GTK_BOX(hbox), label, TRUE, TRUE, 5);
  gtk_widget_show(label);

  /* text entry 1 */
  entry = gtk_entry_new ();
  gtk_box_pack_start (GTK_BOX(hbox), entry, TRUE, TRUE, 5);
  gtk_widget_set_usize (entry, LFG_ENTRY_WIDTH, 0);
  sprintf (buffer, "%d", lfgsvals.xoffset);
  gtk_entry_set_text (GTK_ENTRY(entry), buffer);
  gtk_signal_connect (GTK_OBJECT(entry), "changed",
                      (GtkSignalFunc) lfg_xentry_callback,
                      NULL);
  gtk_widget_show (entry);

  gtk_widget_show(hbox);

  hbox = gtk_hbox_new(FALSE, 5);
  gtk_container_add(GTK_CONTAINER(frame_vbox), hbox);

  /* label 2 */
  label = gtk_label_new("y:");
  gtk_box_pack_start(GTK_BOX(hbox), label, TRUE, TRUE, 5);
  gtk_widget_show(label);
  
  /* button 2 */
  entry = gtk_entry_new ();
  gtk_box_pack_start (GTK_BOX(hbox), entry, TRUE, TRUE, 5);
  gtk_widget_set_usize (entry, LFG_ENTRY_WIDTH, 0);
  sprintf (buffer, "%d", lfgsvals.yoffset);
  gtk_entry_set_text (GTK_ENTRY(entry), buffer);
  gtk_signal_connect (GTK_OBJECT(entry), "changed",
                      (GtkSignalFunc) lfg_yentry_callback,
                      NULL);
  gtk_widget_show (entry);

  gtk_widget_show(hbox);
  gtk_widget_show(frame_vbox);
  gtk_widget_show(frame);

  /* direction */
  frame = gtk_frame_new("Direction");
  gtk_frame_set_shadow_type(GTK_FRAME(frame), GTK_SHADOW_ETCHED_IN);
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG (dlg)->vbox), frame, TRUE, TRUE, 0);

  frame_vbox = gtk_vbox_new(FALSE, 5);
  gtk_container_add(GTK_CONTAINER(frame), frame_vbox);

  /* button 1 */
  button = gtk_radio_button_new_with_label(group, "Vertical");
  group =  gtk_radio_button_group(GTK_RADIO_BUTTON(button)); 
  gtk_box_pack_start(GTK_BOX(frame_vbox), button, TRUE, 0, 0);
  gtk_signal_connect (GTK_OBJECT(button), "toggled",
                      (GtkSignalFunc)toggle_update, &tgl_idx[0]);
  gtk_toggle_button_set_state (GTK_TOGGLE_BUTTON(button), tgl_idx[0]);
  gtk_widget_show(button);

  /* button 2 */
  button = gtk_radio_button_new_with_label(group, "Horizontal");
  group =  gtk_radio_button_group(GTK_RADIO_BUTTON(button)); 
  gtk_box_pack_start(GTK_BOX(frame_vbox), button, TRUE, 0, 0);
  gtk_signal_connect (GTK_OBJECT(button), "toggled",
                      (GtkSignalFunc)toggle_update, &tgl_idx[1]);
  gtk_toggle_button_set_state (GTK_TOGGLE_BUTTON(button), tgl_idx[1]);
  gtk_widget_show(button);

  gtk_widget_show(frame_vbox);
  gtk_widget_show(frame);
  gtk_widget_show (dlg);

  gtk_main ();
  gdk_flush ();
  
  return lfgsint.run;
}

static void save_close_callback (GtkWidget *widget, gpointer data)
{
  gtk_main_quit ();
}

static void save_ok_callback (GtkWidget *widget, gpointer data)
{
  lfgsint.run = TRUE;
  gtk_widget_destroy (GTK_WIDGET (data));
}

static void lfg_xentry_callback (GtkWidget *widget, gpointer data)
{
  lfgsvals.xoffset = atoi(gtk_entry_get_text(GTK_ENTRY(widget)));
}

static void lfg_yentry_callback (GtkWidget *widget, gpointer data)
{
  lfgsvals.yoffset = atoi(gtk_entry_get_text(GTK_ENTRY(widget)));
}

static void toggle_update(GtkWidget *widget, gpointer data)
{
  int *toggle_val;
  
  toggle_val = (int *)data;
  
  if (GTK_TOGGLE_BUTTON(widget)->active) {
    *toggle_val = TRUE;
  } else {
    *toggle_val = FALSE;
  }
}

/*
 * save in LFG format
 */
static void lfg_set_cmap(LFG *lfg, int numcols, guchar *cmap)
{
  int i;

  for (i = 0; i < numcols; i++) {
    lfg->palette[i][R] = optimal_color_component(*cmap++);
    lfg->palette[i][G] = optimal_color_component(*cmap++);
    lfg->palette[i][B] = optimal_color_component(*cmap++);
  }
  for (; i < 16; i++) {
    lfg->palette[i][R] = 0;
    lfg->palette[i][G] = 0;
    lfg->palette[i][B] = 0;
  }
}

/*
 * Each component of RGB can have only 16 variations(00, 11, ..., ff).
 * This function gets a any guchar and returns optimal palette component.
 */
static guchar optimal_color_component(guchar value)
{
  int opt;

  opt = (((int)value + 8) / 0x11) * 0x11;
  return (guchar)opt;
}

static int output_lfg(LFG *lfg, FILE *fp)
{
  int i;
  guchar tmp_pal[48];

  /* output magic */
  fprintf(fp, "LEAFCODE");

  /* output palette */
  for (i = 0; i < 16; i++) {
    tmp_pal[i * 3] = lfg->palette[i][R] >> 4;
    tmp_pal[i * 3 + 1] = lfg->palette[i][G] >> 4;
    tmp_pal[i * 3 + 2] = lfg->palette[i][B] >> 4;
  }
  for (i = 0; i < 24; i++) {
    fputc((tmp_pal[i * 2] << 4) | (tmp_pal[i * 2 + 1]), fp);
  }
  
  /* output image size */
  output_word(lfg->xoffset / 8, fp);
  output_word(lfg->yoffset, fp);
  output_word(lfg->width / 8 - 1, fp);
  output_word(lfg->height - 1, fp);

  /* output direction */
  if (lfg->direction == VERTICAL) {
    fputc(0x00, fp);
  } else {
    fputc(0xff, fp);
  }

  /* output transparent */
  fputc(lfg->transparent, fp);

  /* padding */
  fputc(0x00, fp);
  fputc(0x00, fp);
  
  /* extracted size */
  output_longword(lfg->size, fp);

  /* data body */
  for (i = 0; i < lfg->encoded_size; i++) {
    fputc(lfg->encoded_data[i], fp);
  }
  return 0;
}

static void output_word(int d, FILE *fp)
{
  int d1, d2;

  d1 = d & 0xff;
  d2 = (d >> 8) & 0xff;

  fputc(d1, fp);
  fputc(d2, fp);
}

static void output_longword(int d, FILE *fp)
{
  int d1, d2, d3, d4;

  d1 = d & 0xff;
  d2 = (d >> 8) & 0xff;
  d3 = (d >> 16) & 0xff;
  d4 = (d >> 24) & 0xff;

  fputc(d1, fp);
  fputc(d2, fp);
  fputc(d3, fp);
  fputc(d4, fp);
}

static guchar *pack_pixel(guchar *pixels, int width, int height, int direction)
{
  int i, j, x, y;
  guchar a, b;
  guchar *packed_pixmap;

  packed_pixmap = (guchar *)calloc(width * height / 2, sizeof(guchar));
  if (packed_pixmap == NULL) {
    fprintf(stderr, "LFG: Error - Can't allocate memory.\n");
    return NULL;
  }

  if (direction == VERTICAL) {
    printf("Direction: Vertical\n");
    for (j = 0, x = 0; x < width; x += 2) {
      for (y = 0; y < height; y++) {
	a = pixels[y * width + x];
	b = pixels[y * width + x + 1];
	packed_pixmap[j++] = pack(a, b);
      }
    }
  } else if (direction == HORIZONTAL) {
    printf("Direction: Horizontal\n");
    for (i = 0, j = 0; i < width * height;) {
      a = pixels[i++];
      b = pixels[i++];
      packed_pixmap[j++] = pack(a, b);
    }
  } else {
    /* never reached */
    fprintf(stderr, "LFG: Error - Unexpected error.\n");
  }
  
  return packed_pixmap;
}

static guchar pack(guchar a, guchar b)
{
  guchar a1, b1;

  if (a > 0xf || b > 0xf) {
    fprintf(stderr, "LFG: Error - Palette error.n\n");
  }
  a1 = ((a & 0x8) << 4) | ((a & 0x4) << 3) | ((a & 0x2) << 2) | ((a & 0x1) << 1);
  b1 = ((b & 0x8) << 3) | ((b & 0x4) << 2) | ((b & 0x2) << 1) | (b & 0x1);

  return (a1 | b1);
}

static LFG* init_lfg(void)
{
  LFG *lfg;
  
  lfg = (LFG *)malloc(sizeof(LFG));
  if (lfg == NULL) {
    fprintf(stderr, "LFG: Error - Can't allocate memory.\n");
    exit(1);
  }
  lfg->body = NULL;
  lfg->image = NULL;
  
  return lfg;
}

/*
 * initialize a tree
 */
void init_tree(void)
{
  int i;
  
  for (i = RINGBUF_SIZE + 1; i <= RINGBUF_SIZE + 256; i++) {
    rson[i] = NIL;
  }
  for (i = 0; i < RINGBUF_SIZE; i++) {
    dad[i] = NIL;
  }
}

/*
 * insert a node r into a tree
 */
void insert_node(int r)
{
  int i, p, cmp;
  unsigned char *key;
  
  cmp = 1;
  key = &text[r];
  p = RINGBUF_SIZE + 1 + key[0];
  rson[r] = lson[r] = NIL;
  matchlen = 0;

  for ( ; ; ) {
    if (cmp >= 0) {
      if (rson[p] != NIL) p = rson[p];
      else {  rson[p] = r;  dad[r] = p;  return;  }
    } else {
      if (lson[p] != NIL) p = lson[p];
      else {  lson[p] = r;  dad[r] = p;  return;  }
    }
    for (i = 1; i < F; i++)
      if ((cmp = key[i] - text[p + i]) != 0)  break;
    if (i > matchlen) {
      matchpos = p;
      if ((matchlen = i) >= F)  break;
    }
  }
  dad[r] = dad[p];  lson[r] = lson[p];  rson[r] = rson[p];
  dad[lson[p]] = r;  dad[rson[p]] = r;
  if (rson[dad[p]] == p) {
    rson[dad[p]] = r;
  } else {
    lson[dad[p]] = r;
  }
  /* detach p */
  dad[p] = NIL;
}

/*
 * delete a node p from tree
 */
void delete_node(int p)
{
  int  q;
  
  if (dad[p] == NIL) { /* p isn't found */
    return;
  }
  
  if (rson[p] == NIL) {
    q = lson[p];
  } else if (lson[p] == NIL) {
    q = rson[p];
  } else {
    q = lson[p];
    if (rson[q] != NIL) {
      do {
	q = rson[q];
      } while (rson[q] != NIL);
      rson[dad[q]] = lson[q];
      dad[lson[q]] = dad[q];
      lson[q] = lson[p];
      dad[lson[p]] = q;
    }
    rson[q] = rson[p];
    dad[rson[p]] = q;
  }
  dad[q] = dad[p];
  if (rson[dad[p]] == p) {
    rson[dad[p]] = q;
  } else {
    lson[dad[p]] = q;
  }
  dad[p] = NIL;
}

/*
 * main routine of LFG encoder
 */
void lfg_encode(guchar *img, LFG *lfg)
{
  int i, c, len, r, s, lastmatchlen, codeptr;
  unsigned char code[17], mask;
  unsigned long int incount = 0;
  int img_ptr = 0;
  int lfg_ptr = 0;

  /* initialize a tree */
  init_tree();
  
  code[0] = 0;
  codeptr = 1;
  mask = 0x80;
  s = 0;
  r = RINGBUF_SIZE - F;
  
  /* initialize a ring buffer */
  for (i = s; i < r; i++) {
    text[i] = 0x0;
  }
  for (len = 0; len < F ; len++) {
    c = img[img_ptr++];
    if (img_ptr > lfg->size) break;
    text[r + len] = c;
  }

  incount = len;
  if (incount == 0) {
    return;
  }
  for (i = 1; i <= F; i++) insert_node(r - i);
  insert_node(r);
  
  do {
    if (matchlen > len) {
      matchlen = len;
    }
    if (matchlen < 3) {
      matchlen = 1;
      code[0] |= mask;
      code[codeptr++] = text[r];
    } else {
      code[codeptr++] = (unsigned char)(((matchpos & 0x0f) << 4)| (matchlen - 3));
      code[codeptr++] = (unsigned char)(matchpos >> 4);
    }
    if ((mask >>= 1) == 0) {
      for (i = 0; i < codeptr; i++) {
	lfg->encoded_data[lfg_ptr++] = code[i];
      }
      outcount += codeptr;
      code[0] = 0;
      codeptr = 1;
      mask = 0x80;
    }
    lastmatchlen = matchlen;
    for (i = 0; i < lastmatchlen; i++) {
      c = img[img_ptr++];
      if (img_ptr > lfg->size) {
	break;
      }
      delete_node(s);
      text[s] = c;
      if (s < F - 1) {
	text[s + RINGBUF_SIZE] = c;
      }
      s = (s + 1) & (RINGBUF_SIZE - 1);
      r = (r + 1) & (RINGBUF_SIZE - 1);
      insert_node(r);
    }
    if (img_ptr % 1000 == 0) {
      gimp_progress_update ((double) img_ptr / (double) lfg->size);
    }
    while (i++ < lastmatchlen) {
      delete_node(s);
      s = (s + 1) & (RINGBUF_SIZE - 1);
      r = (r + 1) & (RINGBUF_SIZE - 1);
      if (--len) insert_node(r);
    }
  } while (len > 0);
  
  if (codeptr > 1) {
    for (i = 0; i < codeptr; i++) {
      lfg->encoded_data[lfg_ptr++] = code[i];
    }
    outcount += codeptr;
  }

  lfg->encoded_size = lfg_ptr;
}
